
//	Save the formatter to the prism object for later use
if (!prism.formatter) {
	//	Get the filter
	var filter = prism.$injector.get('$filter');
	//	Create a global
	prism.formatter = {
		// 	Usage: prism.formatter.numeric(value,mask);
		numeric: filter('numeric'),	
		//	Usage: prism.formatter.date(dateObj,mask); OR prism.formatter.date(stringDate,mask);
		datetime: filter('levelDate')	
	};
}

//	Initialize the new widget
prism.registerWidget("TimeSeriesChart", {
	name : "TimeSeriesChart",
	family : "TimeSeries",
	title : "Time Series Chart",
	iconSmall : "/plugins/TimeSeriesChart/TS.png",
	styleEditorTemplate: "/plugins/TimeSeriesChart/styler.html",
	style: {
		showLegend: false,
		showXAxisLabel: false,
		chartType: 'line',
		showStep: false,			
		showMarkers: false,
		groupTooltips: true,
		splitChart: false,
		theme : 'white',
		labelFormat : 't',
		dataLimits: {
			seriesCapacity: 1000000,
			categoriesCapacity: 1000000
		}
	},
	data : {
		selection : [],
		defaultQueryResult : {},	
		panels : [
			{
				name: 'X-Axis',
				type: "visible",
				metadata: {
					types: ['dimensions'],
					maxitems: 1
				},
				allowedDataTypes: ["datetime","numeric"],
				visibility: true
			},
			{
				name: 'value',
				type: 'visible',	
				itemAttributes: ["color"],
				allowedColoringTypes: function() {
					return  {
						color: true,
						condition: false,
						range: false
					};
				},
				metadata: {
					types: ['measures'],
					maxitems: -1
				},
				itemAdded: function(widget, item) {
					var colorFormatType = $$get(item, "format.color.type");
					if ("color" === colorFormatType || "color" === colorFormatType) {
						var color = item.format.color.color;
						defined(color) && "transparent" != color && "white" != color && "#fff" != color && "#ffffff" != color || $jaql.resetColor(item)
					}
					"range" === colorFormatType && $jaql.resetColor(item), defined(item, "format.color_bkp") && $jaql.resetColor(item), defined(item, "format.members") && delete item.format.members
				}
			},
			{
				name: 'breakby',
				label: "Break By",
				type: "series",
				itemAttributes: ["color"],
				allowedColoringTypes: function() {
					return {
						color: true,
						condition: false,
						range: false
					}
				},
				metadata: {
					types: ['dimensions'],
					maxitems: 1
				},
				visibility: true
			},		
			{
				name: 'filters',
				type: 'filters',
				metadata: {
					types: ['dimensions'],
					maxitems: -1
				}
			}
		],
		
		canColor: function (widget, panel, item) {			
			//	Handle the values panel
			if (panel.name == "value") {
				if (widget.metadata.panel("breakby").items.length > 0) {
					return false;
				} else {
					return true
				}
			}
			//	Handle the break by
			if (panel.name == "breakby") {
				return true;
			} else {
				return false;
			}
        },
		
		allocatePanel: function (widget, metadataItem) {			
			// measure
			if (prism.$jaql.isMeasure(metadataItem) && widget.metadata.panel("value").items.length === 0) {

				return "value";
			}
			// dimension
			else if (!prism.$jaql.isMeasure(metadataItem) && widget.metadata.panel("X-Axis").items.length < 3) {

				return "X-Axis";
			}
		},

		// returns true/ reason why the given item configuration is not/supported by the widget
		isSupported: function (items) {

			return this.rankMetadata(items, null, null) > -1;
		},

		// ranks the compatibility of the given metadata items with the widget
		rankMetadata: function (items, type, subtype) {

			var a = prism.$jaql.analyze(items);

			// require 1 measure, 2 dimension
			if (a.measures.length == 2 && a.dimensions.length == 1) {

				return 0;
			}

			return -1;
		},

		// populates the metadata items to the widget
		populateMetadata: function (widget, items) {

			var a = prism.$jaql.analyze(items);

			// allocating dimensions
			widget.metadata.panel("X-Axis").push(a.dimensions);
			widget.metadata.panel("value").push(a.measures);

			// allocating filters
			widget.metadata.panel("filters").push(a.filters);
		},

		// builds a jaql query from the given widget
		buildQuery: function (widget) {
	
			//  Build a jaql query object from widget metadata 
			var query = { 
				datasource: widget.datasource, 
				metadata: [] 
			};

			//  Add Dimensions
			for (i=0; i<widget.metadata.panel("X-Axis").items.length; i++) {
				var newItem = $.extend({},widget.metadata.panel("X-Axis").items[i]);
				//	Filter out nulls
				newItem.jaql.filter = {
					exclude: {
						members: ["N\\A"]
					}
				};				
				query.metadata.push(newItem);			
			}

			//  Break by
			for (i=0; i<widget.metadata.panel("breakby").items.length; i++) {
				query.metadata.push(widget.metadata.panel("breakby").items[i]);
			}

			//  Add Values
			for (i=0; i<widget.metadata.panel("value").items.length; i++) {
				query.metadata.push(widget.metadata.panel("value").items[i]);
			}

			

			//  Add Filters
			widget.metadata.panel('filters').items.forEach(function (item) {
				item = $$.object.clone(item, true);
				item.panel = "scope";
				query.metadata.push(item);
			});

			return query;
		},

		//create highcharts data structure
		processResult : function (widget, queryResult) {		
			
			//	Create array for the new query results
			var newQueryResult = [];

			//	Figure out the data series
			var metadata = queryResult.metadata(),
				hasBreakBy = widget.metadata.panel("breakby").items.length > 0,
				allData = queryResult.$$rows;
			

			//	Get the default colors palette
			var defaultColors = prism.activeDashboard.style.options.palette.colors;

			//	Function to calculate new colors
			function ColorLuminance(hex, lum) {

				// validate hex string
				hex = String(hex).replace(/[^0-9a-f]/gi, '');
				if (hex.length < 6) {
					hex = hex[0]+hex[0]+hex[1]+hex[1]+hex[2]+hex[2];
				}
				lum = lum || 0;

				// convert to decimal and change luminosity
				var rgb = "#", c, i;
				for (i = 0; i < 3; i++) {
					c = parseInt(hex.substr(i*2,2), 16);
					c = Math.round(Math.min(Math.max(0, c + (c * lum)), 255)).toString(16);
					rgb += ("00"+c).substr(c.length);
				}

				return rgb;
			}

			//	Init the dictionary
			var data = {};

			//	Loop through all the data points, and restructure
			for (var k=0; k<allData.length; k++) {

				//	Get the data points
				var dim = allData[k][0],
					valuesIndex = hasBreakBy ? 2 : 1,
					breakbyId = hasBreakBy ? allData[k][valuesIndex-1].text : '';

				//	Loop through all measure values
				for (var l = valuesIndex; l<allData[k].length; l++) {
					
					//	Figure out an id to use
					var seriesId = breakbyId + metadata[l].jaql.dim,
						//seriesName = metadata[l].jaql.title + (hasBreakBy ? ": " + breakbyId : ""),
						seriesName = '',
						measureIndex = l - valuesIndex,
						colorAdjustmentFactor = (-0.25),
						colorAdjustmentValue = Math.ceil(measureIndex/2) * colorAdjustmentFactor;

					//	Determine a good name for the series
					if (hasBreakBy) {
						//	Using a break by, but check how many data series exist
						if (allData[k].length > valuesIndex+1) {
							//	Multiple data series
							seriesName = metadata[l].jaql.title + ": " + breakbyId;
						} else {
							//	Single data series
							seriesName = breakbyId;
						}
					} else {
						//	Not using a break by, use only the measure's title
						seriesName = metadata[l].jaql.title;
					}

					//	Get the data point
					var datapoint = allData[k][l] ? allData[k][l] : {};

					//	color darker for odds
					if (measureIndex % 2 == 1) {
						 colorAdjustmentValue *= (-1);
					}					
					var baseColor = datapoint.color ? datapoint.color : allData[k][valuesIndex-1].color,
						color = (measureIndex == 0) ? baseColor : ColorLuminance(baseColor, colorAdjustmentValue);					
						
					//	Create a data object, if it doesn't exist already
					if (!data[seriesId]){
						data[seriesId] = {
							color: color,
							type: 'line',	//	set this as default, pull from style object on render
							name: seriesName,
							data: []
						}
					}

					//	Handle datetime or ints
					var xValue = (typeof dim.data == "object") ? dim.data.getTime() : dim.data;					

					//	Add the data point
					var point = {
						x: xValue,
						y: datapoint.data,
						name: dim.text,
						text: datapoint.text
					};
					data[seriesId].data.push(point);
				}
			}

			//	Transform into a valid array of series' to provide for highcharts
			for (var series in data) {
			    if (data.hasOwnProperty(series)) {			    	
			        newQueryResult.push(data[series]);
			    }
			}

			return newQueryResult;
		}
	},

	// render
	render : function (widget, event) {

		//	cancel if no results
		if (widget.queryResult.length == 0){
			return;
		}
	
		// 	Get widget element, and clear it out
		var element = $(event.element);
		element.empty();

		var gridStyle = 'dash';
		
		//set the color theme
		var labelColor = 'gray';
		if(widget.style.theme == 'black') {
			element.css('background-color','#040C0C');
			labelColor = 'white';
		}
		else {		
			element.css('background-color','#FFFFFF');
			labelColor = 'gray';		
		}

		//	Get metadata for the chart
		var dimensionMetadata = widget.metadata.panel("X-Axis");
		var dataSeriesMetadata = widget.metadata.panel("value");
		var numSeries = dataSeriesMetadata.items.length;


		//	Get the formatting mask for the x-axis
		var xMask;
		var xLevel;
		var xDatatype = 'numeric';

		if (dimensionMetadata.items[0]){			
			xDatatype = dimensionMetadata.items[0].jaql.datatype;
			xLevel = dimensionMetadata.items[0].jaql.level;
			if(dimensionMetadata.items[0].format) {
				xMask = dimensionMetadata.items[0].format.mask;
			}
		}


		//	Get the formatting mask for the y-axis
		var yMask;
		var yDatatype = 'numeric';

		if (dataSeriesMetadata.items[0]){			
			yDatatype = dataSeriesMetadata.items[0].jaql.datatype;	
			if (dataSeriesMetadata.items[0].format){
				yMask = dataSeriesMetadata.items[0].format.mask;
			}		
		}
		
		

		//	YAxis formatter function
		var axisFormatter = function() {
			//	Get the datatype and init mask and value vars
			var mask,
				value,
				datatype = 'numeric';

			//	Get the mask
			mask = this.axis.options.mask;
			//	Get the value
			value = this.value;
			
			//	Format the value to a text and return
			var text = prism.formatter[datatype](value,mask);			
			return text;
		}

		//	XAxis formatter function
		var xAxisFormatter = function() {
			return epochToDate(this.value);
			/*
			//if numeric then format as date			
			if(xDatatype == 'numeric') {
				return epochToDate(this.value);
			}
			else {
				//	Get the datatype and init mask and value vars
				var mask,
					value,
					datatype = this.axis.options.datatype;

				//	Handle dates differently
				if (datatype == 'datetime'){
					//	Get the mask, for the datetime level
					mask = this.axis.options.mask[this.axis.options.level];
					//	Get the value
					value = new Date(this.value);
				} else {
					//	Get the mask
					mask = this.axis.options.mask;
					//	Get the value
					value = this.value;
				}
				
				//	Format the value to a text and return
				var text = prism.formatter[datatype](value,mask);			
				return text;
			}
			*/
		}

			
		// converts the Unix Time Stamp to a date string
		function epochToDate(n) {
			//formats a unix timestamp as a date string
			//it assumes the time in milliseconds from Jan 01, 1970

			//javascript assumes the time is in GMT so we need to adjust it to the local time
			var gmtDiff = new Date().getTimezoneOffset() * 60000;
    		

			var d = new Date(gmtDiff + parseInt(n, 10));
			var yy = d.getFullYear();
			var mm = 1 + d.getMonth();
			var dd = d.getDate();
			var h = d.getHours();
			var m = d.getMinutes();
			var s = d.getSeconds();
			var ms = d.getMilliseconds();
			var monthNames = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];

			var yys = '' + yy;

			var dateStr = '' + ((mm < 10) ? ('0' + mm) : mm) + '/' + ((dd < 10) ? ('0' + dd) : dd) + '/';
			var hmStr = '' + ((h < 10) ? ('0' + h) : h) + ':' + ((m < 10) ? ('0' + m) : m);
			var secStr = '' + ((s < 10) ? ('0' + s) : s);
			var msStr = '' +  ((ms < 10) ? ('00' + ms) : ((ms < 100) ? '0' + ms : ms));

			var labelStr = '';

			
			if(widget.style.labelFormat == 'd') {
				labelStr = dateStr + yys;
			}
			else if(widget.style.labelFormat == 'dt') {
				labelStr = dateStr + yys.substring(2) + ' ' + hmStr;
			}
			else if(widget.style.labelFormat == 'fd') {
				labelStr =  monthNames[mm -1] + ' ' + ((dd < 10) ? ('0' + dd) : dd) + ', ' + yys;
			}
			else if(widget.style.labelFormat == 'sd') {
				labelStr = monthNames[mm -1] + ' ' + ((dd < 10) ? ('0' + dd) : dd);
			}			
			else if(widget.style.labelFormat == 'fdt') {
				labelStr = monthNames[mm -1] + ' ' + ((dd < 10) ? ('0' + dd) : dd) + ' ' + hmStr;
			}
			else if(widget.style.labelFormat == 'tms') {
				labelStr = hmStr + ':' + secStr + '.' + msStr;
			}
			else if(widget.style.labelFormat == 't') {
				labelStr = hmStr + ':' + secStr;
			}			

			return labelStr;
		}


		//not used
		function numToDate(n) {
			//formats a unix timestamp as HH:mm:ss.ms
			//it assumes the time in milliseconds from Jan 01, 1970
			//javascript assumes the time is in GMT and adjusts it to the local time
			//so we have to manually calculate it


			var hr =  n%86400000;
			var h = Math.floor(hr /3600000);
			var minr = hr%3600000
			var m = Math.floor(minr/60000);
			var secr = minr%60000;
			var s = Math.floor(secr/1000);
			var msr = secr%1000;

			var hh = (h < 10) ? ('0' + h) : '' + h;
			var mm = (m < 10) ? ('0' + m) : '' + m;
			var ss = (s < 10) ? ('0' + s) : '' + s;	

			var msrd = '';
			
			if(msr < 10) {
				msrd = '00' + msr;
			}	
			else if(msr < 100) {
				msrd = '0' + msr;
			}
			else {
				msrd = '' + msr;
			}

			return '' + hh + ':' + mm + ':' + ss + '.' + msrd;
		}

		//formatter function
		
		var numToDateFormatter = function() {
			if(xDatatype == 'numeric') {
				return numToDate(this.value);
			}
			else {
				return axisFormatter();
			}
		}
		

		//formats tooltip data value
		function formatTTValue(seriesName, val) {
			var formattedVal = val;
			if (dataSeriesMetadata.items.length == 1){
				//	Only 1 value
				return prism.formatter['numeric'](val, dataSeriesMetadata.items[0].format.mask)
			} else {
				//	Multiple values, need to find the right format mask
				for(var s = 0; s < numSeries; s++) {
					var st = dataSeriesMetadata.items[s].jaql.title;
					if(st == seriesName) {
						formattedVal = prism.formatter['numeric'](val, dataSeriesMetadata.items[s].format.mask);
						break;
					}		
				}
				return formattedVal;
			}
		}

		//	Tooltip formatter function
		var numToDateFormatterTT = function() {

			var htext = '';
			var header =  '';
			var stable = '';
			var tt = '';
           
           
			if(widget.style.groupTooltips) {
				if(isNaN(this.points[0].x)) {
					htext =  this.points[0].x;
				}
				else {
					htext = epochToDate(this.points[0].x)
				}

				header = '<span style="font-size: 20px">' + htext + '</span><br/>';
				stable = '<table>';
				for(var p = 0; p < this.points.length; p++) {
					stable = stable + '<tr><td><span style="color:' + this.points[p].series.color + '">' + 
					this.points[p].series.name + ' </span></td><td>: </td><td style="text-align:right;width:100px"><b>' + formatTTValue(this.points[p].series.name, this.points[p].y) + '</b></td></tr><br>';
				}

				stable = stable + '</table>';

				tt = header + stable;
			}
			else {
				 tt = '<span style="font-size: 20px">' + epochToDate(this.point.x) + '</span><br><span style="color:' + this.point.series.color + '">' + this.point.series.name +  '</span>: <b>' + this.point.text + '</b><br/>';
			}

			return tt;
		}


		//YAxis template
		var yAxisSub = {
            	title: {
            		style: {
                    	color: labelColor
                	},
            		text: null
            	},
				mask: yMask,
            	datatype: yDatatype,
            	labels: {
            		formatter: axisFormatter,
	                style: {
	                    color: labelColor
	                } 
            	},
				
				gridLineDashStyle: gridStyle
        };

       
		//	Create base highchart options
		var options = {
            chart: {
                zoomType: 'x'
            },            
            xAxis: {
                type: 'linear',
                crosshair: true,
                mask: xMask,
            	datatype: 'numeric',
            	level: xLevel,
            	labels: {
            		formatter: xAxisFormatter,         		
	                style: {
	                    color: labelColor
	                }           		
            	}
            },

			yAxis: {
            	title: {
            		style: {
                    	color: labelColor
                	},
            		text: null
            	},
				mask: yMask,
            	datatype: yDatatype,
            	labels: {
            		formatter: axisFormatter,
	                style: {
	                    color: labelColor
	                } 
            	},
				
				gridLineDashStyle: gridStyle,
				height: '96%',
				top: '4%'
				
            },

            exporting: {
            	enabled: false
            },
            legend: {
            	enabled: widget.style.showLegend
            },
            title: {
            	text: ''
            },
            plotOptions: {
            	series: {
            		step: widget.style.showStep ? 'left' : false,            		
            		marker: {
            			enabled: widget.style.showMarkers
            		}
            	}
            },
            series: widget.queryResult,
            tooltip: {	
            	formatter: numToDateFormatterTT,
    			borderColor: 'gray',
    			shared: widget.style.groupTooltips
    		}
        };

	

        //	Set the X Axis title
        var xTitle = null;
        if (widget.style.showXAxisLabel) {
        	xTitle = dimensionMetadata.items[0] ? dimensionMetadata.items[0].jaql.title : null;
        }
        options.xAxis.title = {
        	text: xTitle
        };  

        //	Set the series type
        var seriesType = widget.style.chartType ? widget.style.chartType : 'line';
		for (var i=0; i<widget.queryResult.length; i++){
			options.series[i].type = seriesType;
			
			//jsd: 2018-02-21 set the data point limit in turbo mode
			options.series[i].turboThreshold = 1000000;
			
			var sr = options.series[i];
			
			var dataLabelsTemplate =  {
                enabled : false
            };
					
			//jsd: 2018-10-19 remove labels, which seem to be the default when moving to 7.1
			for(var d = 0; d < sr.data.length; d++) {
				sr.data[d].dataLabels = dataLabelsTemplate;
			}
			
			//console.log('>>Time Series Chart Series Length: ' + sr.data.length);
		}


		//split the chart based on each item in the value panel
		if(widget.style.splitChart) {

			options.yAxis = [];
			
			var yHeight = 98 / numSeries;
			
			//initial gap on top
			var yTop = 2;
			
			//gap between each axis
			var yGap = 2;
			
			//adjust for the gap
			yHeight = yHeight - yGap;

			for(var v = 0; v < numSeries; v++) {

				//	Get the formatting mask for the y-axis,
				if (dataSeriesMetadata.items[v]) {			
					yDatatype = dataSeriesMetadata.items[v].jaql.datatype;	
					if (dataSeriesMetadata.items[v].format) {
						yMask = dataSeriesMetadata.items[v].format.mask;
					}		
				}

				//split the series
				for(var s = 0; s < widget.queryResult.length; s++) {
					if(widget.queryResult[s].name && widget.queryResult[s].name.indexOf(dataSeriesMetadata.items[v].jaql.title) == 0) {
						//assign the y axis
						options.series[s].yAxis = v;
					}
				}


				//clone the YAxis template
				var yAxisSubClone = $$.object.clone(yAxisSub, true);
				
				
				yAxisSubClone.title.text = dataSeriesMetadata.items[v].jaql.title;
				
				//set the format mask for the YAxis
				yAxisSubClone.mask = yMask;

							
				// jsd: 2018-10-19 set the position and height as % of plot area
				// an option available in the new version of highcharts
				if(numSeries == (v + 1)) {
					//set the last series at the bottom, touching the XAxis
					yAxisSubClone.top = '' + (100 - yHeight) + '%';
				}
				else {
					yAxisSubClone.top = '' + (yTop + (yGap * v ) + (yHeight * v)) + '%';
				}
				
				yAxisSubClone.height = '' + yHeight + '%';
				
				options.yAxis.push(yAxisSubClone);
				
			}
		}
		
		element.highcharts(options);

	},


	destroy : function (s, e) {}
});